import importlib
import inspect
import os
import threading
from typing import Callable

import scrapy
import scrapydo
from fastapi import FastAPI
from scrapy.settings import Settings
from uvicorn import Config, Server


def thread(fn: Callable, params: set) -> threading.Thread:
    t = threading.Thread(
        target=fn,
        daemon=True,
        args=params,
    )

    t.start()

    return t


def threads(count: int, fn: Callable, params: set) -> list:
    threads = []

    for _ in range(count):
        threads.append(thread(fn, params))

    return threads


def discover_spiders(settings: Settings) -> dict:
    """
    tries to discover the available spiders via the provided project settings
    """

    spiders = {}

    def is_spider(obj):
        return inspect.isclass(obj) and issubclass(obj, scrapy.Spider)

    for spiders_module in settings.get("SPIDER_MODULES"):
        parent_mod = importlib.import_module(spiders_module)
        for filename in os.listdir(os.path.dirname(parent_mod.__file__)):
            if not filename.startswith("_"):
                mod_name = filename.split('.')[0]
                mod = importlib.import_module(
                    spiders_module + '.' + mod_name
                )
                for _, v in inspect.getmembers(mod, is_spider):
                    spiders[mod_name] = v

    return spiders


def crawl(spider: scrapy.Spider, settings: scrapy.settings.Settings, args: dict) -> scrapy.crawler.Crawler:
    crawler = scrapydo.run_spider(
        spider,
        settings=settings,
        return_crawler=True,
        capture_items=True,
        **args
    )

    return crawler



import json
import time
import uuid

from fastapi import APIRouter, Request, Response

from . import utils

router = APIRouter()


@router.get('/', description="returns some information about the engine such as the available spiders and the queue backlog count")
async def index(req: Request):
    return {
        'success': True,
        'message': 'under your service, sir :)',
        'payload': {
            'spiders': [spider_name for spider_name in req.app.x.spiders],
            'backlog': req.app.redis.llen(req.app.x.queue_name + '.BACKLOG')
        }
    }


@router.get('/run/{spider_name}', description="execute the specified spider in `{spider_name}` and wait for it to return its result, P.S: any query param will be passed to the spider as argument `-a key=value`")
@router.post('/run/{spider_name}', description="execute the specified spider in `{spider_name}` and wait for it to return its result, P.S: any query param and json post data will be passed to the spider as argument `-a key=value`")
async def run(spider_name: str, req: Request, res: Response):
    args = req.query_params._dict
    spider = req.app.x.spiders.get(spider_name, None)

    if not spider:
        res.status_code = 404
        return {
            'success': False,
            'error': 'no valid spider specified'
        }

    try:
        post_data = await req.json()
        if isinstance(post_data, dict):
            args = {**args, **post_data}
    except:
        pass

    result = utils.crawl(spider, req.app.x.settings, args)
    items = result.items
    stats = result.stats.get_stats()

    if stats.get('spider_exceptions/Exception', 0) > 0 and len(items) < 1:
        res.status_code = 500
        return {
            'success': False,
            'error': 'something went wrong, there may be an exception in your spider request',
            'payload': {
                'stats': stats,
                'items': items,
            }
        }

    return {
        'success': True,
        'payload': {
            'items': items,
            'stats': stats,
        }
    }


@router.get('/enqueue/{spider_name}', description="adding the specified spider in `{spider_name}` to the backlog to be executed later, P.S: any query param will be used as spider argument")
@router.post('/enqueue/{spider_name}', description="adding the specified spider in `{spider_name}` to the backlog to be executed later, P.S: any query param and json post data will be used as spider argument")
async def enqueue(spider_name: str, req: Request, res: Response):
    args = req.query_params._dict
    spider = req.app.x.spiders.get(spider_name, None)

    if not spider:
        res.status_code = 404
        return {
            'success': False,
            'error': 'invalid spider specified'
        }

    try:
        post_data = await req.json()
        if isinstance(post_data, dict):
            args = {**args, **post_data}
    except:
        pass

    args["created_at"] = int(time.time())
    args["jobid"] = str(uuid.uuid4())

    task = {
        'spider': spider_name,
        'args': args,
    }

    req.app.redis.rpush(
        req.app.x.queue_name + ".BACKLOG",
        json.dumps(task)
    )

    return {
        'success': True,
        'task': task,
    }



import asyncio
import json
import logging
import os
import sys

import coloredlogs
import redis
import scrapydo
from fastapi import FastAPI
from scrapy.commands import ScrapyCommand
from scrapy.utils.project import get_project_settings
from uvicorn import Config, Server

from . import routes, utils


class Command(ScrapyCommand):
    requires_project = True
    default_settings = {'LOG_ENABLED': False}

    def __init__(self):
        command_name = os.path.basename(__file__).split('.')[0]

        if len(sys.argv) < 2:
            return

        if sys.argv[1] is not command_name:
            return

        self.boot()

        if len(self.settings.get('SPIDER_MODULES', [])) < 1 or len(self.spiders) < 1:
            self.logger.critical(
                'there is no need to call me here, I cannot find any project/spider there!')
            return

        utils.threads(self.queue_workers_count, self.consumer, ())
        loop = asyncio.get_event_loop()

        self.server(loop)

    def boot(self):
        self.logger = logging.getLogger("scrapy-x")

        scrapydo.setup()
        coloredlogs.install(
            fmt="[%(levelname)s] | %(asctime)s |  %(message)s",
            logger=self.logger
        )

        self.settings = get_project_settings()
        self.queue_name = self.settings.get('X_QUEUE_NAME', 'SCRAPY_X_QUEUE')

        self.queue_workers_count = self.settings.getint(
            'X_QUEUE_WORKERS_COUNT', os.cpu_count()
        )

        self.server_workers_count = self.settings.getint(
            'X_SERVER_WORKERS_COUNT', os.cpu_count()
        )

        self.server_listen_port = self.settings.getint(
            'X_SERVER_LISTEN_PORT', 6800
        )

        self.server_listen_host = self.settings.get(
            'X_SERVER_LISTEN_HOST', '0.0.0.0'
        )

        self.enable_access_log = self.settings.getbool(
            'X_ENABLE_ACCESS_LOG', True
        )

        self.redis_config = {
            'host': self.settings.get('X_REDIS_HOST', 'localhost'),
            'port': self.settings.getint('X_REDIS_PORT', 6379),
            'db': self.settings.getint('X_REDIS_DB', 0),
            'password': self.settings.get('X_REDIS_PASSWORD', ''),
        }

        self.spiders = utils.discover_spiders(self.settings)

    def consumer(self):
        """
        start a single redis consumer worker
        """

        try:
            r = redis.Redis(
                host=self.redis_config["host"],
                port=self.redis_config["port"],
                password=self.redis_config["password"],
                db=self.redis_config["db"]
            )
        except Exception as e:
            self.logger.critical("[redis] {}".format(str(e)))
            exit(e)

        while True:
            _, payload = r.blpop(self.queue_name + ".BACKLOG")

            try:
                task = json.loads(payload)
            except Exception as e:
                self.logger.error("invalid task payload {}".format(str(e)))
                continue

            spider_name = task.get("spider", None)
            spider = self.spiders.get(spider_name, None)
            args = task.get("args", {})

            if not isinstance(args, dict):
                self.logger.warning(
                    "invalid args object, replacing it with empty one {}".format(
                        args)
                )

                args = {}

            if not spider:
                self.logger.error("unknwon spider {}".format(spider_name))
                continue

            try:
                utils.crawl(spider, self.settings, args)
            except Exception as e:
                self.logger.critical(e)
                exit(e)

    def server(self, loop):
        """
            prepare the server and initialize it
            so it be ready to run inside a thread

            this code is inspired by the following link:
                - https://github.com/encode/uvicorn/issues/706#issuecomment-652220153
        """

        app = FastAPI()
        app.x = self
        app.redis = redis.Redis(
            host=self.redis_config["host"],
            port=self.redis_config["port"],
            password=self.redis_config["password"],
            db=self.redis_config["db"]
        )

        app.include_router(routes.router)

        config = Config(
            app=app,
            loop=loop,
            workers=self.server_workers_count,
            port=self.server_listen_port,
            host=self.server_listen_host,
            access_log=self.enable_access_log,
            debug=False
        )

        server = Server(config)

        loop.run_until_complete(server.serve())

    def short_desc(self):
        return "start the x server and queue manager"

    def run(self, opts, args):
        pass



from setuptools import find_packages, setup

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open("README.rst", "r", encoding="utf-8") as f:
    README = f.read()

setup(
    name='scrapy-x',
    packages=find_packages(),
    install_requires=required,
    version='1.0.2',
    author='Mohamed Al Ashaal',
    author_email='m7medalash3al@gmail.com',
    license='Apache License V2.0',
    description='a scrapy subcommand for easily enqueuing crawling jobs in a scalable and high performance way',
    summary='a scrapy serving module',
    url='https://github.com/alash3al/scrapyx',
    python_requires='>=3.6.9',
    entry_points={
        'scrapy.commands': [
            'x=scrapyx.x:Command',
        ],
    },
    long_description=README,
)
